unit InstFunc;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Misc. installation functions
}

interface

uses
  WinProcs, WinTypes, SysUtils {$IFNDEF WIN32}, Ver {$ENDIF};

{$I VERSION.INC}

type
  PSimpleStringListArray = ^TSimpleStringListArray;
  {$IFDEF WIN32}
  TSimpleStringListArray = array[0..$1FFFFFFE] of String;
  {$ELSE}
  TSimpleStringListArray = array[0..$3FFE] of PString;
  {$ENDIF}
  TSimpleStringList = class
  private
    FList: PSimpleStringListArray;
    FCount, FCapacity: Integer;
    function Get (Index: Integer): String;
    procedure SetCapacity (NewCapacity: Integer);
  public
    destructor Destroy; override;
    procedure Add (const S: String);
    procedure AddIfDoesntExist (const S: String);
    procedure Clear;

    property Count: Integer read FCount;
    property Items[Index: Integer]: String read Get; default;
  end;

type
  TDeleteDirProc = function(const DirName: String; Param: Pointer): Boolean;

const
  {$IFDEF WIN32}
  RegRootKeyNames: array[HKEY_CLASSES_ROOT..HKEY_LOCAL_MACHINE] of PChar =
    ('HKEY_CLASSES_ROOT', 'HKEY_CURRENT_USER', 'HKEY_LOCAL_MACHINE');
  {$ELSE}
  RegClassesRootName = 'HKEY_CLASSES_ROOT';
  {$ENDIF}

function DelTree (const Path: String; const IsDir, DeleteFiles, DeleteSubdirsAlso: Boolean;
  const DeleteDirProc: TDeleteDirProc; const Param: Pointer): Boolean;
function InstExec (const Filename, Params: String; WorkingDir: String;
  const WaitUntilTerminated, WaitUntilIdle: Boolean;
  const ProcessMessagesProc: TProcedure; var ErrorCode: Integer): Boolean;
function InstShellExec (const Filename, Params: String; WorkingDir: String;
  var ErrorCode: Integer): Boolean;
function GenerateUniqueName (Path: String; const Extension: String): String;
function GetVersionInfo ({$IFDEF WIN32}const{$ENDIF} Filename: String;
  var VersionInfo: {$IFDEF WIN32} TVSFixedFileInfo {$ELSE} tvs_FixedFileInfo {$ENDIF}): Boolean;
procedure RestartReplace ({$IFDEF WIN32} const {$ENDIF} TempFile, DestFile: String);
{$IFDEF WIN32}
function GetFileDateTime (const Filename: string; var DateTime: TFileTime): Boolean;
function GetFileLocalDateTime (const Filename: string; var DateTime: TFileTime): Boolean;
procedure IncrementSharedCount (const Filename: String;
  const AlreadyExisted: Boolean);
{$ELSE}
function ThunkGetVersion: Longint;
{$ENDIF}

implementation

uses
  ShellApi,
  CmnFunc2,
  Msgs, MsgIDs {$IFNDEF WIN32} , Thunk {$ENDIF};

function GenerateUniqueName (Path: String; const Extension: String): String;
  function IntToBase32 (Number: Longint): String;
  const
    Table: array[0..31] of Char = '0123456789ABCDEFGHIJKLMNOPQRSTUV';
  var
    I: Integer;
  begin
    Result := '';
    for I := 0 to 4 do begin
      Insert (Table[Number and 31], Result, 1);
      Number := Number shr 5;
    end;
  end;
var
  Rand, RandOrig: Longint;
begin
  Path := AddBackslash(Path);
  {$IFDEF WIN32}
  RandOrig := Random($2000000);
  {$ELSE}
  LongRec(RandOrig).Lo := Random($FFFF);
  LongRec(RandOrig).Hi := Random($200);
  {$ENDIF}
  Rand := RandOrig;
  repeat
    Inc (Rand);
    if Rand > $1FFFFFF then Rand := 0;
    if Rand = RandOrig then
      { practically impossible to go through 33 million possibilities,
        but check "just in case"... }
      raise Exception.Create(FmtSetupMessage1(msgErrorTooManyFilesInDir,
        RemoveBackslashUnlessRoot(Path)));
    { Generate a random name }
    Result := Path + 'is-' + IntToBase32(Rand) + Extension;
  until not FileOrDirExists(Result);
end;

procedure RestartReplace ({$IFDEF WIN32} const {$ENDIF} TempFile, DestFile: String);
{ Renames TempFile to DestFile the next time Windows is started. If DestFile
  already existed, it will be overwritten. If DestFile is '' then TempFile
  will be deleted, however this is only supported by 95/98 and NT, not
  Windows 3.1x. }
var
  WinDir, WinInitFile, TempWinInitFile: String;
  OldF, NewF: TextFile;
  OldFOpened, NewFOpened: Boolean;
  L, L2: String;
  RenameSectionFound, WriteLastLine: Boolean;
  NewDestFile: PChar;
{$IFNDEF WIN32}
  Kernel32Inst: DWORD;
  idMoveFileEx: DWORD;
  MoveFileEx: function(lpExistingFileName, lpNewFileName: PChar; dwFlags: DWORD;
    thunkProcAddress, thunkAddressConvert, thunkNParams: DWORD): LongBool
    absolute CallProc32W;
  idGetLastError: DWORD;
  GetLastError: function(thunkProcAddress, thunkAddressConvert,
    thunkNParams: DWORD): DWORD absolute CallProc32W;
const
  MOVEFILE_REPLACE_EXISTING   = $00000001;
  MOVEFILE_DELAY_UNTIL_REBOOT = $00000004;
  procedure ThunkError (const Code: Integer);
  begin
    raise Exception.Create(FmtSetupMessage1(msgErrorThunk, IntToStr(Code)));
  end;
{$ENDIF}
begin
  if not UsingWinNT then begin
    { Because WININIT.INI allows multiple entries with the same name,
      it must manually parse the file instead of using
      WritePrivateProfileString }
    WinDir := GetWinDir;
    WinInitFile := AddBackslash(WinDir) + 'WININIT.INI';
    OldFOpened := False;
    NewFOpened := False;
    try
      try
        if FileExists(WinInitFile) then begin
          TempWinInitFile := GenerateUniqueName(WinDir, '.tmp');
          { Flush Windows' cache for the file first }
          WritePrivateProfileString (nil, nil, nil, StringAsPChar(WinInitFile));
          AssignFile (OldF, WinInitFile);
          FileMode := fmOpenRead or fmShareDenyWrite;  Reset (OldF);
          OldFOpened := True;
        end
        else
          TempWinInitFile := WinInitFile;
        AssignFile (NewF, TempWinInitFile);
        FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (NewF);
        NewFOpened := True;
        RenameSectionFound := False;
        WriteLastLine := False;
        if OldFOpened then
          while not Eof(OldF) do begin
            Readln (OldF, L);
            WriteLastLine := True;
            L2 := Trim(L);
            if (L2 <> '') and (L2[1] = '[') then begin
              if CompareText(L, '[rename]') = 0 then
                RenameSectionFound := True
              else
              if RenameSectionFound then
                Break;
            end;
            Writeln (NewF, L);
            WriteLastLine := False;
          end;
        if not RenameSectionFound then
          Writeln (NewF, '[rename]');
        if DestFile <> '' then
          L2 := GetShortName(DestFile)
        else
          L2 := 'NUL';
        Writeln (NewF, L2 + '=' + GetShortName(TempFile));
        if OldFOpened then begin
          if WriteLastLine then
            Writeln (NewF, L);
          while not Eof(OldF) do begin
            Readln (OldF, L);
            Writeln (NewF, L);
          end;
        end;
      finally
        if NewFOpened then CloseFile (NewF);
        if OldFOpened then CloseFile (OldF);
      end;
    except
      if OldFOpened then DeleteFile (TempWinInitFile);
      raise;
    end;
    if OldFOpened then begin
      AssignFile (OldF, WinInitFile);
      Erase (OldF);
      AssignFile (OldF, TempWinInitFile);
      Rename (OldF, WinInitFile);
    end;
  end
  else begin
    {$IFNDEF WIN32}
    if not ThunkingFunctionsPresent then
      ThunkError (0);
    Kernel32Inst := LoadLibraryEx32W('kernel32.dll', 0, 0);
    if Kernel32Inst = 0 then
      ThunkError (1);
    try
      idMoveFileEx := GetProcAddress32W(Kernel32Inst, 'MoveFileExA');
      if idMoveFileEx = 0 then
        ThunkError (10);
      idGetLastError := GetProcAddress32W(Kernel32Inst, 'GetLastError');
      if idGetLastError = 0 then
        ThunkError (11);
    {$ENDIF}
      if DestFile <> '' then
        NewDestFile := StringAsPChar(DestFile)
      else
        NewDestFile := nil;
      if not MoveFileEx(StringAsPChar(TempFile), NewDestFile,
         MOVEFILE_DELAY_UNTIL_REBOOT or MOVEFILE_REPLACE_EXISTING
         {$IFNDEF WIN32} , idMoveFileEx, 6, 3 {$ENDIF}) then
        raise Exception.Create(FmtSetupMessage(msgErrorFunctionFailed,
          ['MoveFileEx', IntToStr(GetLastError {$IFNDEF WIN32} (idGetLastError, 0, 0) {$ENDIF})]));
    {$IFNDEF WIN32}
    finally
      FreeLibrary32W (Kernel32Inst);
    end;
    {$ENDIF}
  end;
end;

function DelTree (const Path: String; const IsDir, DeleteFiles, DeleteSubdirsAlso: Boolean;
  const DeleteDirProc: TDeleteDirProc; const Param: Pointer): Boolean;
{ Deletes the specified directory including all files and subdirectories in
  it (including those with hidden, system, and read-only attributes). Returns
  True if it was able to successfully remove everything. }
{$IFDEF WIN32}
var
  BasePath, FindSpec: String;
  H: THandle;
  FindData: TWin32FindData;
  S: String;
begin
  Result := True;
  if DeleteFiles then begin
    if IsDir then begin
      BasePath := AddBackslash(Path);
      FindSpec := BasePath + '*';
    end
    else begin
      BasePath := ExtractFilePath(Path);
      FindSpec := Path;
    end;
    H := FindFirstFile(PChar(FindSpec), FindData);
    if H <> INVALID_HANDLE_VALUE then begin
      repeat
        S := FindData.cFileName;
        if FindData.dwFileAttributes and FILE_ATTRIBUTE_READONLY <> 0 then
          SetFileAttributes (PChar(BasePath + S), FindData.dwFileAttributes and
            not FILE_ATTRIBUTE_READONLY);
        if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0 then
          WinProcs.DeleteFile (PChar(BasePath + S))
        else begin
          if DeleteSubdirsAlso and (S <> '.') and (S <> '..') then
            if not DelTree(BasePath + S, True, True, True, DeleteDirProc, Param) then
              Result := False;
        end;
      until not FindNextFile(H, FindData);
      WinProcs.FindClose (H);
    end;
  end;
  if IsDir then begin
    if Assigned(DeleteDirProc) then begin
      if not DeleteDirProc(Path, Param) then
        Result := False;
    end
    else begin
      if not RemoveDirectory(PChar(Path)) then
        Result := False;
    end;
  end;
end;
{$ELSE}
var
  BasePath, FindSpec: String;
  SR: TSearchRec;
begin
  Result := True;
  if DeleteFiles then begin
    if IsDir then begin
      BasePath := AddBackslash(Path);
      FindSpec := BasePath + '*.*';
    end
    else begin
      BasePath := ExtractFilePath(Path);
      FindSpec := Path;
    end;
    if FindFirst(FindSpec, faHidden or faSysFile or faDirectory, SR) = 0 then begin
      repeat
        if SR.Attr and faReadOnly <> 0 then
          FileSetAttr (BasePath + SR.Name, SR.Attr and not faReadOnly);
        if SR.Attr and faDirectory = 0 then
          DeleteFile (BasePath + SR.Name)
        else begin
          if DeleteSubdirsAlso and (SR.Name <> '.') and (SR.Name <> '..') then
            if not DelTree(BasePath + SR.Name, True, True, True, DeleteDirProc, Param) then
              Result := False;
        end;
      until FindNext(SR) <> 0;
      FindClose (SR);
    end;
  end;
  if IsDir then begin
    if Assigned(DeleteDirProc) then begin
      if not DeleteDirProc(Path, Param) then
        Result := False;
    end
    else begin
      try
        RmDir (Path);
      except
        Result := False;
      end;
    end;
  end;
end;
{$ENDIF}

function GetVersionInfo ({$IFDEF WIN32}const{$ENDIF} Filename: String;
  var VersionInfo: {$IFDEF WIN32} TVSFixedFileInfo {$ELSE} tvs_FixedFileInfo {$ENDIF}): Boolean;
var
  VersionSize: Integer;
  VersionHandle: DWORD;
  VersionBuf: PChar;
  VerInfo: {$IFDEF WIN32} PVSFixedFileInfo {$ELSE} pvs_FixedFileInfo {$ENDIF};
  VerInfoSize: UINT;
begin
  Result := False;

  VersionSize := GetFileVersionInfoSize(StringAsPChar(Filename),
    VersionHandle);
  if VersionSize <> 0 then begin
    GetMem (VersionBuf, VersionSize);
    try
      if GetFileVersionInfo(StringAsPChar(Filename), VersionHandle, VersionSize, VersionBuf) then begin
        if VerQueryValue(VersionBuf, '\', Pointer(VerInfo), VerInfoSize) then begin
          VersionInfo := VerInfo^;
          Result := True;
        end;
      end;
    finally
      FreeMem (VersionBuf, VersionSize);
    end;
  end;
end;

{$IFDEF WIN32}
procedure IncrementSharedCount (const Filename: String;
  const AlreadyExisted: Boolean);
const
  SharedDLLsKey = NEWREGSTR_PATH_SETUP + '\SharedDLLs';  {don't localize}
var
  K: HKEY;
  Disp, Size, Count, CurType, NewType: DWORD;
  CountStr: String;
  FilenameP: PChar;
begin
  if RegCreateKeyEx(HKEY_LOCAL_MACHINE, SharedDLLsKey, 0, nil, REG_OPTION_NON_VOLATILE,
     KEY_QUERY_VALUE or KEY_SET_VALUE, nil, K, @Disp) <> ERROR_SUCCESS then
    raise Exception.Create(FmtSetupMessage(msgErrorRegOpenKey,
      [RegRootKeyNames[HKEY_LOCAL_MACHINE], SharedDLLsKey]));
  FilenameP := PChar(Filename);
  Count := 0;
  NewType := REG_DWORD;
  try
    if RegQueryValueEx(K, FilenameP, nil, @CurType, nil, @Size) = ERROR_SUCCESS then
      case CurType of
        REG_SZ:
          if RegQueryStringValue(K, FilenameP, CountStr) then begin
            Count := StrToInt(CountStr);
            NewType := REG_SZ;
          end;
        REG_BINARY: begin
            if (Size >= 1) and (Size <= 4) then begin
              if RegQueryValueEx(K, FilenameP, nil, nil, @Count, @Size) <> ERROR_SUCCESS then
                { ^ relies on the high 3 bytes of Count being initialized to 0 }
                Abort;
              NewType := REG_BINARY;
            end;
          end;
        REG_DWORD: begin
            Size := SizeOf(DWORD);
            if RegQueryValueEx(K, FilenameP, nil, nil, @Count, @Size) <> ERROR_SUCCESS then
              Abort;
          end;
      end;
  except
    Count := 0;
  end;
  if Integer(Count) < 0 then Count := 0;  { just in case... }
  if (Count = 0) and AlreadyExisted then
    Inc (Count);
  Inc (Count);
  case NewType of
    REG_SZ: begin
        CountStr := IntToStr(Count);
        RegSetValueEx (K, FilenameP, 0, NewType, PChar(CountStr), Length(CountStr)+1);
      end;
    REG_BINARY, REG_DWORD:
      RegSetValueEx (K, FilenameP, 0, NewType, @Count, SizeOf(Count));
  end;
  RegCloseKey (K);
end;
{$ENDIF}

{$IFNDEF WIN32}
function ThunkGetVersion: Longint;
{ 16-bit only function: Thunks to kernel32.dll to get the Win32 subsystem
  version number. If the OS doesn't support generic thunks (i.e. Windows 3.1)
  it returns the Win16 version number. }
var
  Kernel32Inst: DWORD;
  idGetVersion32: DWORD;
begin
  if ThunkingFunctionsPresent then begin
    Kernel32Inst := LoadLibraryEx32W('kernel32.dll', 0, 0);
    if Kernel32Inst <> 0 then
      try
        idGetVersion32 := GetProcAddress32W(Kernel32Inst, 'GetVersion');
        if idGetVersion32 <> 0 then begin
          Result := CallProc32W(idGetVersion32, 0, 0);
          Exit;
        end;
      finally
        FreeLibrary32W (Kernel32Inst);
      end;
  end;
  Result := GetVersion;
end;
{$ENDIF}

{$IFDEF WIN32}
function GetFileDateTime (const Filename: string; var DateTime: TFileTime): Boolean;
var
  Handle: THandle;
  FindData: TWin32FindData;
begin
  Handle := FindFirstFile(PChar(Filename), FindData);
  if Handle <> INVALID_HANDLE_VALUE then begin
    WinProcs.FindClose (Handle);
    if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0 then begin
      DateTime := FindData.ftLastWriteTime;
      Result := True;
      Exit;
    end;
  end;
  Result := False;
  DateTime.dwLowDateTime := 0;
  DateTime.dwHighDateTime := 0;
end;

function GetFileLocalDateTime (const Filename: string; var DateTime: TFileTime): Boolean;
var
  DT: TFileTime;
begin
  Result := GetFileDateTime(Filename, DT);
  if Result then
    FileTimeToLocalFileTime (DT, DateTime);
end;
{$ENDIF}

function InstExec (const Filename, Params: String; WorkingDir: String;
  const WaitUntilTerminated, WaitUntilIdle: Boolean;
  const ProcessMessagesProc: TProcedure; var ErrorCode: Integer): Boolean;
var
  CmdLine: String;
  {$IFDEF WIN32}
  WorkingDirP: PChar;
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  {$ELSE}
  FilenameP: array[0..255] of Char;
  E: Integer;
  SaveCurDir: String;
  {$ENDIF}
begin
  Result := True;
  CmdLine := {$IFDEF WIN32}AddQuotes({$ENDIF}Filename{$IFDEF WIN32}){$ENDIF} +
    ' ' + Params {$IFNDEF WIN32} + #0 {needs null terminator for Delphi 1} {$ENDIF};
  if WorkingDir = '' then
    WorkingDir := RemoveBackslashUnlessRoot(ExtractFilePath(Filename));
  {$IFDEF WIN32}
  FillChar (StartupInfo, SizeOf(StartupInfo), 0);
  StartupInfo.cb := SizeOf(StartupInfo);
  if WorkingDir <> '' then
    WorkingDirP := PChar(WorkingDir)
  else
    WorkingDirP := nil;
  if not CreateProcess(nil, PChar(CmdLine), nil, nil, False, 0, nil,
     WorkingDirP, StartupInfo, ProcessInfo) then begin
    Result := False;
    ErrorCode := GetLastError;
    Exit;
  end;
  with ProcessInfo do begin
    { Don't need the thread handle, so close it now }
    CloseHandle (hThread);
    if WaitUntilIdle then
      WaitForInputIdle (hProcess, INFINITE);
    if WaitUntilTerminated then
      { Wait until the process returns, but still process any messages that
        arrive. }
      repeat
        { Process any pending messages first because MsgWaitForMultipleObjects
          (called below) only returns when *new* messages arrive }
        if Assigned(ProcessMessagesProc) then
          ProcessMessagesProc;
      until MsgWaitForMultipleObjects(1, hProcess, False, INFINITE, QS_ALLINPUT) <> WAIT_OBJECT_0+1;
    { Then close the process handle }
    CloseHandle (hProcess);
  end;
  {$ELSE}
  SaveCurDir := GetCurrentDir;
  try
    if WorkingDir <> '' then
      SetCurrentDir (WorkingDir);
    E := WinExec(@CmdLine[1], SW_SHOW);
    if E < 32 then begin
      Result := False;
      ErrorCode := E;
      Exit;
    end;
  finally
    SetCurrentDir (SaveCurDir);
  end;
  if WaitUntilTerminated then begin
    { Wait until the process returns }
    StrPCopy (FilenameP, Filename);
    while GetModuleHandle(FilenameP) <> 0 do
      if Assigned(ProcessMessagesProc) then
        ProcessMessagesProc;
  end;
  {$ENDIF}
end;

function InstShellExec (const Filename, Params: String; WorkingDir: String;
  var ErrorCode: Integer): Boolean;
var
  WorkingDirP: PChar;
  E: Integer;
  {$IFNDEF WIN32}
  Filename2, Params2, WorkingDir2: array[0..255] of Char;
  SaveCurDir: String;
  {$ENDIF}
begin
  if WorkingDir = '' then
    WorkingDir := RemoveBackslashUnlessRoot(ExtractFilePath(Filename));
  {$IFDEF WIN32}
  if WorkingDir <> '' then
    WorkingDirP := PChar(WorkingDir)
  else
    WorkingDirP := nil;
  E := ShellExecute(0, 'open', PChar(Filename), PChar(Params), WorkingDirP,
    SW_SHOW);
  {$ELSE}
  StrPCopy (Filename2, Filename);
  StrPCopy (Params2, Params);
  StrPCopy (WorkingDir2, Params);
  if WorkingDir2[0] <> #0 then
    WorkingDirP := WorkingDir2
  else
    WorkingDirP := nil;
  E := ShellExecute(0, 'open', Filename2, Params2, WorkingDirP, SW_SHOW);
  {$ENDIF}
  Result := E > 32;
  if not Result then
    ErrorCode := E;
end;


{ TSimpleStringList }

procedure TSimpleStringList.Add (const S: String);
begin
  if FCount = FCapacity then
    SetCapacity (FCapacity + 8);
  {$IFDEF WIN32}
  FList^[FCount] := S;
  {$ELSE}
  FList^[FCount] := NewStr(S);
  {$ENDIF}
  Inc (FCount);
end;

procedure TSimpleStringList.AddIfDoesntExist (const S: String);
var
  I: Integer;
begin
  for I := 0 to FCount-1 do
    if FList^[I]{$IFNDEF WIN32}^{$ENDIF} = S then
      Exit;
  Add (S);
end;

procedure TSimpleStringList.SetCapacity (NewCapacity: Integer);
begin
  {$IFDEF WIN32}
  ReallocMem (FList, NewCapacity * SizeOf(Pointer));
  if NewCapacity > FCapacity then
    FillChar (FList^[FCapacity], (NewCapacity - FCapacity) * SizeOf(Pointer), 0);
  {$ELSE}
  FList := ReallocMem(FList, FCapacity * SizeOf(Pointer),
    NewCapacity * SizeOf(Pointer));
  {$ENDIF}
  FCapacity := NewCapacity;
end;

procedure TSimpleStringList.Clear;
{$IFNDEF WIN32}
var
  I: Integer;
{$ENDIF}
begin
  {$IFDEF WIN32}
  if FCount <> 0 then Finalize (FList^[0], FCount);
  {$ELSE}
  for I := FCount-1 downto 0 do
    DisposeStr (FList^[I]);
  {$ENDIF}
  FCount := 0;
  SetCapacity (0);
end;

function TSimpleStringList.Get (Index: Integer): String;
begin
  Result := FList^[Index]{$IFNDEF WIN32}^{$ENDIF};
end;

destructor TSimpleStringList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

end.
